﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace Parade.Metadata.Asf
{
	/// <summary>
	/// ASF ファイルのタグ情報を編集します。
	/// </summary>
	public class AsfTagEditor : IDisposable
	{
		/// <summary>
		/// インスタンスを初期化します。
		/// </summary>
		/// <param name="fileName">ASF ファイル。</param>
		/// <exception cref="ArgumentNullException">fileName が null 参照です。</exception>
		/// <exception cref="ArgumentException">fileName が空の文字列 ("") か、空白しか含んでいないか、無効な文字を 1 つ以上含んでいます。またはファイル以外のデバイスを参照しています ( NTFS 環境の "con:"、"com1:"、"lpt1:" など )。</exception>
		/// <exception cref="NotSupportedException">fileName がファイル以外のデバイスを参照しています (NTFS 以外の環境の "con:"、"com1:"、"lpt1:" など)。</exception>
		/// <exception cref="FileNotFoundException">ファイルが見つかりません。</exception>
		/// <exception cref="SecurityException">呼び出し元に、必要なアクセス許可がありません。</exception>
		/// <exception cref="PathTooLongException">指定したパス、ファイル名、またはその両方がシステム定義の最大長を超えています。たとえば、Windows ベースのプラットフォームの場合、パスの長さは 248 文字未満、ファイル名の長さは 260 文字未満である必要があります。</exception>
		/// <exception cref="NotSupportedException">ASF ファイルではありません。</exception>
		public AsfTagEditor( string fileName )
		{
			var reader = new FileStream( fileName, FileMode.Open, FileAccess.Read, FileShare.Read );
			this.Load( reader );

			this._reader = reader;
			this._fileName = fileName;
		}

		/// <summary>
		/// リソースを破棄します。
		/// </summary>
		public void Dispose()
		{
			if( this._reader != null )
			{
				this._reader.Dispose();
				this._reader = null;
			}

			if( this._tags != null )
			{
				this._tags.Clear();
				this._tags = null;
			}

			GC.SuppressFinalize( this );
		}

		/// <summary>
		/// AST のタグ情報をダンプします。
		/// </summary>
		/// <param name="filePath">ASF ファイルへのパス情報。</param>
		/// <param name="dumpFilePath">ダンプ先となるテキストファイルへのパス情報。</param>
		/// <exception cref="NotSupportedException">ASF ファイルではありません。</exception>
		public static void Dump( string filePath, string dumpFilePath )
		{
			using( var dumper = new TagDumper( filePath, dumpFilePath ) )
			{
				dumper.Dump();
			}
		}

		/// <summary>
		/// タグ情報を読み込みます。
		/// </summary>
		/// <param name="reader">ASF のタグ情報を読み取るためのストリーム。</param>
		/// <exception cref="NotSupportedException">ASF ファイルではありません。</exception>
		private void Load( Stream reader )
		{
			// 書き込みを終えたあと、再初期化のために呼び出されることを想定しておく
			this._tags.Clear();
			reader.Seek( 0, SeekOrigin.Begin );

			// ASF ファイルであることをチェック
			{
				var info = new HeaderObjectInfo( reader );
				if( info.Guid != Guids.HeaderObject ) { throw new NotSupportedException( "ASF file format is not." ); }
			}

			// オブジェクトの総数取得と Reserved 1、2 のスキップ
			var count = reader.ReadInt32();
			reader.Seek( 2, SeekOrigin.Current );

			// オブジェクト読み取り
			for( var i = 0; i < count; ++i )
			{
				var top = reader.Position;
				var info = new HeaderObjectInfo( reader );

				if( info.Guid == Guids.ContentDescriptionObject )
				{
					this.LoadContentDescriptionObject( reader );
				}
				else if( info.Guid == Guids.ExtendedContentDescriptionObject )
				{
					this.LoadExtendedContentDescriptionObject( reader );
				}
				else if( info.Guid == Guids.FilePropertiesObject )
				{
					this.LoadFilePropertiesObject( reader );
				}

				// 次のオブジェクト始点へ移動
				reader.Seek( top + info.Size, SeekOrigin.Begin );
			}
		}

		/// <summary>
		/// Content Description Object の内容を読み取ります。
		/// </summary>
		/// <param name="reader">ASF のタグ情報を読み取るためのストリーム。</param>
		private void LoadContentDescriptionObject( Stream reader )
		{
			var lengths = new int[ 5 ] { reader.ReadInt16(), reader.ReadInt16(), reader.ReadInt16(), reader.ReadInt16(), reader.ReadInt16() };
			var names   = new string[ 5 ] { AsfDefinedTags.Title.Name, AsfDefinedTags.Author.Name, AsfDefinedTags.Copyright.Name, AsfDefinedTags.Description.Name, AsfDefinedTags.Rating.Name };

			for( var i = 0; i < 5; ++i )
			{
				this._tags.Add( names[ i ], new TagData( reader.Position, lengths[ i ], TagDataType.String ) );
				reader.Seek( lengths[ i ], SeekOrigin.Current );
			}
		}

		/// <summary>
		/// Extended Content Description Object の内容を読み取ります。
		/// </summary>
		/// <param name="reader">ASF のタグ情報を読み取るためのストリーム。</param>
		private void LoadExtendedContentDescriptionObject( Stream reader )
		{
			ushort count = reader.ReadUInt16();
			for( ushort i = 0; i < count; ++i )
			{
				var name   = reader.ReadStringUtf16( reader.ReadUInt16() );
				var type   = ( TagDataType )reader.ReadUInt16();
				var length = reader.ReadUInt16();

				this._tags.Add( name, new TagData( reader.Position, length, type ) );
				reader.Seek( length, SeekOrigin.Current );
			}
		}

		/// <summary>
		/// File Properties Object の内容をダンプします。
		/// </summary>
		/// <param name="reader">ASF のタグ情報を読み取るためのストリーム。</param>
		/// <param name="writer">タグ情報を出力するためのストリーム。</param>
		private void LoadFilePropertiesObject( Stream reader )
		{
			// File ID をスキップ
			reader.Seek( 16, SeekOrigin.Current );

			// File Size
			this._tags.Add( AsfDefinedTags.FileSize.Name, new TagData( reader.Position, 8, TagDataType.UInt64 ) );
			reader.Seek( 8, SeekOrigin.Current );

			// Creation Date ( 8 )、Data Packets Count ( 8 ) をスキップ
			reader.Seek( 16, SeekOrigin.Current );

			// Play Duration
			var duration = reader.ReadUInt64();

			// Send Duration をスキップ
			reader.Seek( 8, SeekOrigin.Current );

			// Play Duration から Preroll x 10000 を差し引いたものが演奏時間となる。
			// この値は複数値からの算出が必要なので、遅延読み込みせずにキャッシュする。
			//
			duration -= ( reader.ReadUInt64() * 10000 );
			this._tags.Add( AsfDefinedTags.Duration.Name, new TagData( 0, 0, TagDataType.UInt64, duration ) );

			// ... 現時点では他のデータは不要
		}

		/// <summary>
		/// タグ情報を読み取ります。
		/// </summary>
		/// <param name="name">タグ名。</param>
		/// <returns>成功時はタグ情報。それ以外は null 参照。</returns>
		public object Read( string name )
		{
			// 未定義のタグは対象外
			var info = AsfDefinedTags.GetInfo( name );
			if( info == null ) { return null; }

			TagData data;
			if( !this._tags.TryGetValue( name, out data ) ) { return null; }
			
			return this.Read( info.Type, data );
		}

		/// <summary>
		/// タグ情報を削除します。
		/// </summary>
		/// <param name="name">タグ名。</param>
		/// <returns>成功時はタグ情報。それ以外は null 参照。</returns>
		/// <exception cref="NotSupportedException">サポートされていない操作をおこなおうとしました。</exception>
		private void Remove( string name )
		{
			// 未実装
			throw new NotImplementedException();

			/*
			TagData data;
			if( !this._tags.TryGetValue( name, out data ) || data.IsRemove ) { throw new NotSupportedException( "Remove tag was not found." ); }

			data.IsRemove = true;
			*/
		}

		/// <summary>
		/// タグ情報を読み取ります。
		/// </summary>
		/// <param name="type">タグのデータ型。</param>
		/// <param name="data">タグ情報。</param>
		/// <returns>成功時はタグ情報。それ以外は null 参照。</returns>
		private object Read( AsfTagDataType type, TagData data )
		{
			// 書き込み候補、またはキャッシュ済みのデータが存在するなら優先する
			if( data.Value != null ) { return data.Value; }

			var reader = this._reader;
			reader.Seek( data.Position, SeekOrigin.Begin );

			// 外部公開する値とファイルのデータ型に生じる齟齬を解決する。
			// 数値文字列、またはその逆など、相互変換が可能なものであればサポートする。
			//
			switch( type )
			{
			case AsfTagDataType.Binary:
				if( data.Type == TagDataType.Binary ) { return reader.ReadByteArray( data.Length ); }
				break;

			case AsfTagDataType.Bool:
				switch( data.Type )
				{
				case TagDataType.Bool:   return reader.ReadBool();
				case TagDataType.UInt16: return ( reader.ReadUInt16() != 0 );
				case TagDataType.UInt32: return ( reader.ReadUInt32() != 0 );
				case TagDataType.UInt64: return ( reader.ReadUInt64() != 0 );
				}
				break;

			case AsfTagDataType.Guid:
				switch( data.Type )
				{
				case TagDataType.Binary: return reader.ReadGuid();
				case TagDataType.String: return new Guid( reader.ReadStringUtf16( data.Length ) );
				}
				break;

			case AsfTagDataType.String:
				switch( data.Type )
				{
				case TagDataType.Bool:   return reader.ReadBool().ToString();
				case TagDataType.String: return reader.ReadStringUtf16( data.Length );
				case TagDataType.UInt16: return reader.ReadUInt16().ToString();
				case TagDataType.UInt32: return reader.ReadUInt32().ToString();
				case TagDataType.UInt64: return reader.ReadUInt64().ToString();
				}
				break;

			case AsfTagDataType.UInt16:
				switch( data.Type )
				{
				case TagDataType.String:
					{
						ushort result;
						if( ushort.TryParse( reader.ReadStringUtf16( data.Length ), out result ) ) { return result; }
					}
					break;
				
				case TagDataType.UInt16: return reader.ReadUInt16();
				case TagDataType.UInt32: return ( ushort )reader.ReadUInt32();
				case TagDataType.UInt64: return ( ushort )reader.ReadUInt64();
				}
				break;

			case AsfTagDataType.UInt32:
				switch( data.Type )
				{
				case TagDataType.String:
					{
						uint result;
						if( uint.TryParse( reader.ReadStringUtf16( data.Length ), out result ) ) { return result; }
					}
					break;
				
				case TagDataType.UInt16: return ( uint )reader.ReadUInt16();
				case TagDataType.UInt32: return reader.ReadUInt32();
				case TagDataType.UInt64: return ( uint )reader.ReadUInt64();
				}
				break;

			case AsfTagDataType.UInt64:
				switch( data.Type )
				{
				case TagDataType.String:
					{
						ulong result;
						if( ulong.TryParse( reader.ReadStringUtf16( data.Length ), out result ) ) { return result; }
					}
					break;

				case TagDataType.UInt16: return ( ulong )reader.ReadUInt16();
				case TagDataType.UInt32: return ( ulong )reader.ReadUInt32();
				case TagDataType.UInt64: return reader.ReadUInt64();
				}
				break;
			}

			return null;
		}

		/// <summary>
		/// 編集内容を保存します
		/// </summary>
		public void Save()
		{
			// 未実装
			throw new NotImplementedException();
		}

		/// <summary>
		/// タグ情報を書き込みます。
		/// 既存のタグがなければ追加、ある場合は更新をおこないます。
		/// </summary>
		/// <param name="name">タグ名。</param>
		/// <param name="value">タグ情報。値を削除する場合は null を指定します。</param>
		/// <exception cref="NotSupportedException">サポートされていない操作をおこなおうとしました。</exception>
		/// <exception cref="ArgumentException"></exception>
		public void Write( string name, object value )
		{
			// 未実装
			throw new NotImplementedException();

			/*
			// 未定義のタグは対象外
			var info = AsfDefinedTags.GetInfo( name );
			if( info == null ) { throw new NotSupportedException( "Tag is not defined." ); }

			if( info.IsReadOnly ) { throw new NotSupportedException( "Tag is read-only." ); }

			// 削除
			if( value == null )
			{
				this.Remove( name );
				return;
			}

			// 型チェック
			if( info.Type.IsValidType( value ) ) { throw new ArgumentException( "Invalid value type." ); }

			TagData data;
			if( this._tags.TryGetValue( name, out data ) )
			{
				data.Value    = value;
				data.IsRemove = false;
			}
			else
			{
				data = new TagData( 0, 0, info.Type.ToTagDataType(), value );
				this._tags.Add( name, data );
			}
			*/
		}

		/// <summary>
		/// ASF ファイル。
		/// </summary>
		private string _fileName;

		/// <summary>
		/// ASF ファイルへの読み取り用ストリーム。
		/// </summary>
		private Stream _reader;

		/// <summary>
		/// タグ名をキーとする、タグ情報のディクショナリ。
		/// </summary>
		private Dictionary< string, TagData > _tags = new Dictionary< string, TagData >();
	}
}
